// Stage3D Shoot-em-up Tutorial Part 4
// by Christer Kaitila - www.mcfunkypants.com

// EntityManager.as
// The entity manager handles a list of all known game entities.
// This object pool will allow for reuse (respawning) of
// sprites: for example, when enemy ships are destroyed,
// they will be re-spawned when needed as an optimization 
// that increases fps and decreases ram use.

package
{
	import flash.display.Bitmap;
	import flash.display3D.*;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	public class EntityManager
	{
		// the level data parser
		public var level:GameLevels;
		// the current level number
		public var levelNum:int = 0;
		// where in the level we are in pixels
		public var levelCurrentScrollX:Number = 0;
		// the last spawned column of level data
		public var levelPrevCol:int = -1;
		// pixels we need to scroll before spawning the next col
		public var levelTilesize:int = 48;
		// this is used to ensure all terrain tiles line up exactly
		public var lastTerrainEntity:Entity;
		
		// we need to allow at least enough space for ship movement
		// entities that move beyond the edges of the screen
		// plus this amount are recycled (destroyed for reuse)
		public var cullingDistance:Number = 200;
		
		// a particle system class that updates our sprites
		public var particles:GameParticles;
		
		// so that explosions can be played
		public var sfx:GameSound;
		
		// the sprite sheet image
		public var spriteSheet : LiteSpriteSheet;
		public var SpritesPerRow:int = 8;
		public var SpritesPerCol:int = 8;
		// we no longer force a particular spritesheet here
		//[Embed(source="../assets/sprites.png")]
		//private var SourceImage : Class;
		public var SourceImage : Class;

		// the general size of the player and enemies
		public var defaultScale:Number = 1; // 1.5;
		// how fast the default scroll (enemy flying) speed is
		public var defaultSpeed:Number = 128;
		// how fast player bullets go per second
		public var bulletSpeed:Number = 250;

		// for framerate-independent timings
		public var currentFrameSeconds:Number = 0;
		
		// sprite IDs (indexing the spritesheet)
		public const spritenumFireball:uint = 63;
		public const spritenumFireburst:uint = 62;
		public const spritenumShockwave:uint = 61;
		public const spritenumDebris:uint = 60;
		public const spritenumSpark:uint = 59;
		public const spritenumBullet3:uint = 58;
		public const spritenumBullet2:uint = 57;
		public const spritenumBullet1:uint = 56;
		public const spritenumPlayer:uint = 10;
		public const spritenumOrb:uint = 17;

		// reused for calculation speed
		public const DEGREES_TO_RADIANS:Number = Math.PI / 180;
		public const RADIANS_TO_DEGREES:Number = 180 / Math.PI;
		
		// the player entity - a special case
		public var thePlayer:Entity;
		// a "power orb" that orbits the player
		public var theOrb:Entity;
		
		// a reusable pool of entities
		// this contains every known Entity
		// including the contents of the lists below
		public var entityPool : Vector.<Entity>;
		// these pools contain only certain types
		// of entity as an optimization for smaller loops 
		public var allBullets : Vector.<Entity>;
		public var allEnemies : Vector.<Entity>;
		
		// all the polygons that make up the scene
		public var batch : LiteSpriteBatch;
		
		// for statistics
		public var numCreated : int = 0;
		public var numReused : int = 0;
		
		public var maxX:int;
		public var minX:int;
		public var maxY:int;
		public var minY:int;
		public var midpoint:int;
		
		public function EntityManager(view:Rectangle)
		{
			entityPool = new Vector.<Entity>();
			allBullets = new Vector.<Entity>();
			allEnemies = new Vector.<Entity>();
			particles = new GameParticles(this);
			setPosition(view);	
			level = new GameLevels();
		}
		
		public function setPosition(view:Rectangle):void 
		{
			// allow moving fully offscreen before
			// automatically being culled (and reused)
			maxX = view.width + cullingDistance;
			minX = view.x - cullingDistance;
			maxY = view.height + cullingDistance;
			minY = view.y - cullingDistance;
			midpoint = view.height / 2;
		}
		
		// this XOR based fast random number generator runs 4x faster
		// than Math.random() and also returns a number from 0 to 1
		// see http://www.calypso88.com/?cat=7
		private const FASTRANDOMTOFLOAT:Number = 1 / uint.MAX_VALUE;
		private var fastrandomseed:uint = Math.random() * uint.MAX_VALUE;
		public function fastRandom():Number
		{
			fastrandomseed ^= (fastrandomseed << 21);
			fastrandomseed ^= (fastrandomseed >>> 35);
			fastrandomseed ^= (fastrandomseed << 4);
			return (fastrandomseed * FASTRANDOMTOFLOAT);
		}

		public function createBatch(context3D:Context3D, uvPadding:Number=0) : LiteSpriteBatch 
		{
			var sourceBitmap:Bitmap = new SourceImage();

			// create a spritesheet with 8x8 (64) sprites on it
			spriteSheet = new LiteSpriteSheet(sourceBitmap.bitmapData, SpritesPerRow, SpritesPerCol, uvPadding);
			
			// Create new render batch 
			batch = new LiteSpriteBatch(context3D, spriteSheet);
			
			return batch;
		}
		
		// search the entity pool for unused entities and reuse one
		// if they are all in use, create a brand new one
		public function respawn(sprID:uint=0):Entity
		{
			var currentEntityCount:int = entityPool.length;
			var anEntity:Entity;
			var i:int = 0;
			// search for an inactive entity
			for (i = 0; i < currentEntityCount; i++ ) 
			{
				anEntity = entityPool[i];
				if (!anEntity.active && (anEntity.sprite.spriteId == sprID))
				{
					//trace('Reusing Entity #' + i);
					anEntity.active = true;
					anEntity.sprite.visible = true;
					anEntity.recycled = true;
					numReused++;
					return anEntity;
				}
			}
			// none were found so we need to make a new one
			//trace('Need to create a new Entity #' + i);
			var sprite:LiteSprite;
			sprite = batch.createChild(sprID);
			anEntity = new Entity(sprite, this);
			entityPool.push(anEntity);
			numCreated++;
			return anEntity;
		}
		
		// this entity is the PLAYER
		public function addPlayer(playerController:Function):Entity 
		{
			trace("Adding Player Entity");
			thePlayer = respawn(spritenumPlayer);
			thePlayer.sprite.position.x = 64;
			thePlayer.sprite.position.y = midpoint;
			thePlayer.sprite.rotation = 180 * DEGREES_TO_RADIANS;
			thePlayer.sprite.scaleX = thePlayer.sprite.scaleY = defaultScale; 
			thePlayer.speedX = 0;
			thePlayer.speedY = 0;
			thePlayer.active = true;
			thePlayer.collidemode = 1;
			thePlayer.collideradius = 10;
			thePlayer.owner = thePlayer; // collisions require this
			thePlayer.aiFunction = playerController;
			
			// just for fun, spawn an orbiting "power orb"
			theOrb = respawn(spritenumOrb);
			theOrb.rotationSpeed = 720 * DEGREES_TO_RADIANS;
			theOrb.sprite.scaleX = theOrb.sprite.scaleY = defaultScale / 2;
			theOrb.leavesTrail = true;
			theOrb.collidemode = 1;
			theOrb.collideradius = 12;
			theOrb.isBullet = true;
			theOrb.owner = thePlayer;
			theOrb.orbiting = thePlayer;
			theOrb.orbitingDistance = 180;
			
			return thePlayer;
		}		

		// shoot a bullet
		public function shootBullet(powa:uint=1, shooter:Entity = null):Entity 
		{
			// just in case the AI is running during the main menu
			// and we've not yet created the player entity
			if (thePlayer == null) return null;

			var theBullet:Entity;
			// assume the player shot it
			// otherwise maybe an enemy did
			if (shooter == null) 
				shooter = thePlayer;
				
			// three possible bullets, progressively larger
			if (powa == 1) 
				theBullet = respawn(spritenumBullet1);
			else if (powa == 2) 
				theBullet = respawn(spritenumBullet2);
			else 
				theBullet = respawn(spritenumBullet3);
			theBullet.sprite.position.x = shooter.sprite.position.x + 8;
			theBullet.sprite.position.y = shooter.sprite.position.y + 2;
			theBullet.sprite.rotation = 180 * DEGREES_TO_RADIANS;
			theBullet.sprite.scaleX = theBullet.sprite.scaleY = 1; 
			if (shooter == thePlayer)
			{
				theBullet.speedX = bulletSpeed;
				theBullet.speedY = 0;
			}
			else // enemy bullets move slower and towards the player
			{
				theBullet.sprite.rotation = 
					pointAtRad(theBullet.sprite.position.x - thePlayer.sprite.position.x,
						theBullet.sprite.position.y-thePlayer.sprite.position.y) - (90*DEGREES_TO_RADIANS);
				
				// move in the direction we're facing
				theBullet.speedX = defaultSpeed*1.5*Math.cos(theBullet.sprite.rotation);
				theBullet.speedY = defaultSpeed*1.5*Math.sin(theBullet.sprite.rotation);
   
				// optionally, we could just fire straight ahead in the direction we're heading:
				// theBullet.speedX = shooter.speedX * 1.5;
				// theBullet.speedY = shooter.speedY * 1.5;
				// and we could point where we're going like this:
				// pointAtRad(theBullet.speedX,theBullet.speedY) - (90*DEGREES_TO_RADIANS);
			}
			theBullet.owner = shooter;
			theBullet.collideradius = 10;
			theBullet.collidemode = 1;
			theBullet.isBullet = true;
			if (!theBullet.recycled)
				allBullets.push(theBullet);
			return theBullet;
		}		

		// Unused: this was "addEntities()" in the previous tutorials.
		// It spawns random enemies that move in a straight line
		public function addRandomEntity():void
		{
			var anEntity:Entity;
			var sprID:int;
			sprID = Math.floor(fastRandom() * 55);
			// try to reuse an inactive entity (or create a new one)
			anEntity = respawn(sprID);
			// give it a new position and velocity
			anEntity.sprite.position.x = maxX;
			anEntity.sprite.position.y = fastRandom() * maxY;
			anEntity.speedX = 15 * ((-1 * fastRandom() * 10) - 2);
			anEntity.speedY = 15 * ((fastRandom() * 5) - 2.5);
			anEntity.sprite.scaleX = defaultScale;
			anEntity.sprite.scaleY = defaultScale;
			anEntity.sprite.rotation = pointAtRad(anEntity.speedX,anEntity.speedY) - (90*DEGREES_TO_RADIANS);
			anEntity.collidemode = 1;
			anEntity.collideradius = 16;
			if (!anEntity.recycled)
				allEnemies.push(anEntity);
		}
		
		// returns the angle in radians of two points
		public function pointAngle(point1:Point, point2:Point):Number
		{
			var dx:Number = point2.x - point1.x;
			var dy:Number = point2.y - point1.y;
			return -Math.atan2(dx,dy);
		}		
		
		// returns the angle in degrees of 0,0 to x,y
		public function pointAtDeg(x:Number, y:Number):Number
		{
			return -Math.atan2(x,y) * RADIANS_TO_DEGREES;
		}		

		// returns the angle in radians of 0,0 to x,y
		public function pointAtRad(x:Number, y:Number):Number
		{
			return -Math.atan2(x,y);
		}		

		// as an optimization to saver millions of checks, only
		// the player's bullets check for collisions with all enemy ships
		// (enemy bullets only check to hit the player)
		public function checkCollisions(checkMe:Entity):Entity
		{
			var anEntity:Entity;
			var collided:Boolean = false;
			if (checkMe.owner != thePlayer)
			{	// quick check ONLY to see if we have hit the player
				anEntity = thePlayer;
				if (checkMe.colliding(anEntity)) 
				{
					trace("Player was HIT!");
					collided = true;
				}
			}
			else // check all active enemies
			{
				for(var i:int=0; i< allEnemies.length;i++)
				{
					anEntity = allEnemies[i];
					if (anEntity.active && anEntity.collidemode)
					{
						if (checkMe.colliding(anEntity)) 
						{
							collided = true;
							break;
						}
					}
				}
			}
			if (collided)
			{
				//trace('Collision!');
				if (sfx) sfx.playExplosion(int(fastRandom() * 2 + 1.5));
				particles.addExplosion(checkMe.sprite.position);
				if ((checkMe != theOrb) && (checkMe != thePlayer)) 
					checkMe.die(); // the bullet
				if ((anEntity != theOrb) && ((anEntity != thePlayer))) 
					anEntity.die(); // the victim
				return anEntity;
			}
			return null;
		}
		
		// called every frame: used to update the simulation
		// this is where you would perform AI, physics, etc.
		// in this version, currentTime is seconds since the previous frame
		public function update(currentTime:Number) : void
		{		
			var anEntity:Entity;
			var i:int;
			var max:int;
			
			// what portion of a full second has passed since the previous update?
			currentFrameSeconds = currentTime / 1000;
			
			// handle all other entities
			max = entityPool.length;
			for (i = 0; i < max; i++)
			{
				anEntity = entityPool[i];
				if (anEntity.active)
				{
					// subtract the previous aiPathOffset
					anEntity.sprite.position.x -= anEntity.aiPathOffsetX;
					anEntity.sprite.position.y -= anEntity.aiPathOffsetY;

					// calculate location on screen with scrolling
					anEntity.sprite.position.x += anEntity.speedX * currentFrameSeconds;
					anEntity.sprite.position.y += anEntity.speedY * currentFrameSeconds;
										
					// is a custom AI specified? if so, run it now
					if (anEntity.aiFunction != null)
					{
						anEntity.aiFunction(currentFrameSeconds);
					}

					// add the new aiPathOffset
					anEntity.sprite.position.x += anEntity.aiPathOffsetX;
					anEntity.sprite.position.y += anEntity.aiPathOffsetY;
					
					// collision detection
					if (anEntity.collidemode) // && anEntity.isBullet
					{
						checkCollisions(anEntity);
					}
					
					// entities can orbit other entities 
					// (uses their rotation as the position)
					if (anEntity.orbiting != null)
					{
						anEntity.sprite.position.x = anEntity.orbiting.sprite.position.x + 
							((Math.sin(anEntity.sprite.rotation/4)/Math.PI) * anEntity.orbitingDistance);
						anEntity.sprite.position.y = anEntity.orbiting.sprite.position.y - 
							((Math.cos(anEntity.sprite.rotation/4)/Math.PI) * anEntity.orbitingDistance);
					}

					// entities can leave an engine emitter trail
					if (anEntity.leavesTrail)
					{
						// leave a trail of particles
						if (anEntity == theOrb)
							particles.addParticle(63, 
							anEntity.sprite.position.x, anEntity.sprite.position.y, 
							0.25, 0, 0, 0.6, NaN, NaN, -1.5, -1);
						else // other enemies
							particles.addParticle(63, anEntity.sprite.position.x + 12, 
							anEntity.sprite.position.y + 2, 
							0.5, 3, 0, 0.6, NaN, NaN, -1.5, -1);
						
					}
					
					if ((anEntity.sprite.position.x > maxX) ||
						(anEntity.sprite.position.x < minX) ||
						(anEntity.sprite.position.y > maxY) ||
						(anEntity.sprite.position.y < minY))							
					{
						// if we go past any edge, become inactive
						// so the sprite can be respawned
						if ((anEntity != thePlayer) && (anEntity != theOrb)) 
							anEntity.die();
					}
					
					if (anEntity.rotationSpeed != 0)
						anEntity.sprite.rotation += anEntity.rotationSpeed * currentFrameSeconds;
						
					if (anEntity.fadeAnim != 0)
					{
						anEntity.sprite.alpha += anEntity.fadeAnim * currentFrameSeconds;
						if (anEntity.sprite.alpha <= 0.001)
						{
							anEntity.die();
						}
						else if (anEntity.sprite.alpha > 1)
						{
							anEntity.sprite.alpha = 1;
						}
					}
					if (anEntity.zoomAnim != 0)
					{
						anEntity.sprite.scaleX += anEntity.zoomAnim * currentFrameSeconds;
						anEntity.sprite.scaleY += anEntity.zoomAnim * currentFrameSeconds;
						if (anEntity.sprite.scaleX < 0 || anEntity.sprite.scaleY < 0)
							anEntity.die();
					}
				}
			}
		}

		// kill (recycle) all known entities
		// this is run when we change levels
		public function killEmAll():void
		{
			//trace('Killing all entities...');
			var anEntity:Entity;
			var i:int;
			var max:int;
			max = entityPool.length;
			for (i = 0; i < max; i++)
			{
				anEntity = entityPool[i];
				if ((anEntity != thePlayer) && (anEntity != theOrb))
					anEntity.die();
			}
		}
		
		// load a new level for entity generation
		public function changeLevels(lvl:String):void
		{
			killEmAll();
			level.loadLevel(lvl);
			levelCurrentScrollX = 0;
			levelPrevCol = -1;
		}
		
		// check to see if another row from the level data should be spawned 
		public function streamLevelEntities(theseAreEnemies:Boolean = false):void 
		{
			var anEntity:Entity;
			var sprID:int;
			// time-based with overflow remembering (increment and floor)
			levelCurrentScrollX += defaultSpeed * currentFrameSeconds;
			// is it time to spawn the next col from our level data?
			if (levelCurrentScrollX >= levelTilesize)
			{
				levelCurrentScrollX = 0;
				levelPrevCol++;
				
				// this prevents small "seams" due to floating point inaccuracies over time
				var currentLevelXCoord:Number;
				if (lastTerrainEntity && !theseAreEnemies) 
					currentLevelXCoord = lastTerrainEntity.sprite.position.x + levelTilesize;
				else
					currentLevelXCoord = maxX;
				
				var rows:int = level.data.length;
				//trace('levelCurrentScrollX = ' + levelCurrentScrollX + 
				//' - spawning next level column ' + levelPrevCol + ' row count: ' + rows);
								
				if (level.data && level.data.length)
				{
					for (var row:int = 0; row < rows; row++)
					{
						if (level.data[row].length > levelPrevCol) // data exists? NOP?
						{
							//trace('Next row data: ' + String(level.data[row]));
							sprID = level.data[row][levelPrevCol];
							if (sprID > -1) // zero is a valid number, -1 means blank
							{
								anEntity = respawn(sprID);
								anEntity.sprite.position.x = currentLevelXCoord;
								anEntity.sprite.position.y = (row * levelTilesize) + (levelTilesize/2);
								trace('Spawning a level sprite ID ' + sprID + ' at ' 
									+ anEntity.sprite.position.x + ',' + anEntity.sprite.position.y);
								anEntity.speedX = -defaultSpeed;
								anEntity.speedY = 0;
								anEntity.sprite.scaleX = defaultScale;
								anEntity.sprite.scaleY = defaultScale;
								
								if (theseAreEnemies)
								{
									// which AI should we give this enemy?
									switch (sprID)
									{
										case 1:
										case 2:
										case 3:
										case 4:
										case 5:
										case 6:
										case 7:
											// move forward at a random angle
											anEntity.speedX = 15 * ((-1 * fastRandom() * 10) - 2);
											anEntity.speedY = 15 * ((fastRandom() * 5) - 2.5);
											anEntity.aiFunction = anEntity.straightAI;
											break;
										case 8:
										case 9:
										case 10:
										case 11:
										case 12:
										case 13:
										case 14:
										case 15:
											// move straight with a wobble
											anEntity.aiFunction = anEntity.wobbleAI;
											break
										case 16:
										case 24: // sentry guns don't move and always look at the player
											anEntity.aiFunction = anEntity.sentryAI;
											anEntity.speedX = -90; // same speed as background
											break;
										case 17:
										case 18:
										case 19:
										case 20:
										case 21:
										case 22:
										case 23:
											// move at a random angle with a wobble
											anEntity.speedX = 15 * ((-1 * fastRandom() * 10) - 2);
											anEntity.speedY = 15 * ((fastRandom() * 5) - 2.5);
											anEntity.aiFunction = anEntity.wobbleAI;
											break;
										case 32:
										case 40:
										case 48: // asteroids don't move or shoot but they do spin and drift
											anEntity.aiFunction = null;
											anEntity.rotationSpeed = fastRandom() * 8 - 4
											anEntity.speedY = fastRandom() * 64 - 32;
											break;
										default: // follow a complex random spline curve path
											anEntity.aiFunction = anEntity.droneAI;
											break;
									}
									
									anEntity.sprite.rotation = pointAtRad(anEntity.speedX, anEntity.speedY)
										- (90*DEGREES_TO_RADIANS);
									anEntity.collidemode = 1;
									anEntity.collideradius = 16;
									if (!anEntity.recycled)
										allEnemies.push(anEntity);
								} // end if these were enemies
							}// end loop for level data rows
						}
					}
				}
				// remember the last created terrain entity
				// (might be null if the level data was blank for this column)
				// to avoid slight seams due to terrain scrolling speed over time
				if (!theseAreEnemies) lastTerrainEntity = anEntity;
			}
		}
	} // end class
} // end package